from fakts_next.cli.advertise import (
    advertise,
    retrieve_bindings,
    AdvertiseBinding,
    AdvertiseBeacon,
)
import rich_click as click
import asyncio
from rich.console import Console
from rich.prompt import Prompt
from rich.panel import Panel
import sys
from typing import List

console = Console()

logo = """
    __      _   _      
  / _|__ _| |_| |_ ___
 |  _/ _` | / /  _(_-/
 |_| \__,_|_\_\\__/__/                     
 """

welcome = r"\nThe fakts_next beacon lets you advertise a fakts_next endpoint on your local network. This is useful if you want to advertise a fakts_next endpoint to other devices on your local network."


async def adversite_all(
    bindings: List[AdvertiseBinding],
    beacons: List[AdvertiseBeacon],
    interval: int = 1,
    iterations: int = 10,
) -> None:
    """Advertise on all bindings

    This function will advertise on all bindings
    all beacons.  And will wait for all of them to finish.

    Parameters
    ----------
    bindings : List[AdvertiseBinding]
        The bindings to use
    beacons : List[AdvertiseBeacon]
        The beacons / well-known-endpoints to advertise
    interval : int, optional
        How often to advertise (in seconds), by default 1
    iterations : int, optional
        How often should we adversite? If -1 its infiinite, by default 10
    """

    await asyncio.gather(
        *[
            advertise(binding, beacons, interval=interval, iterations=iterations)
            for binding in bindings
        ]
    )


@click.group()
def cli() -> None:
    """
    Fakts cli lets you interact with the fakts_next api
    through the command line.

    """
    pass


@cli.command("beacon", short_help="Advertises a fakts_next endpoint")
@click.option("--url", "-u", help="The url to advertise", required=False)
@click.option("--all", "-a", help="Advertise on all interfaces", is_flag=True)
@click.option(
    "--iterations",
    "-i",
    help="How many iterations (-1 equals indefintetly)",
    default=-1,
    type=int,
)
@click.option(
    "--interval",
    "-i",
    type=int,
    help="Which second interval between broadcast",
    default=5,
)
def beacon(url: str, all: bool, iterations: int, interval: int) -> None:
    """Runs the arkitekt app (using a builder)"""

    md = Panel(logo + welcome, title="Fakts Beacon", title_align="center")
    console.print(md)
    try:
        bindings = retrieve_bindings()
    except ImportError:
        error = Panel(
            """netifaces is required to use the advertised discovery. please install it seperately or install fakts_next with the 'beacon' extras: pip install "fakts_next\[beacon]" """,
            title="Fakts Beacon",
            title_align="center",
            style="red",
        )
        console.print(error)
        sys.exit(1)

    console.print("Which Interface should be used for broadcasting?")

    if not url:
        url = Prompt.ask("What is the url of the endpoint you want to advertise?")

    all = (
        all
        or Prompt.ask(
            "Do you want to use all interfaces?", default="y", choices=["y", "n"]
        )
        == "y"
    )

    console.print(
        f"Advertising {url} every {interval} seconds " + "forever"
        if iterations == -1
        else f"{iterations} times"
    )

    if not all:
        for i, binding in enumerate(bindings):
            console.print(
                f"[{i}] : Use interface {binding.interface}: {binding.bind_addr} advertising to {binding.broadcast_addr}"
            )

        bind_index = Prompt.ask(
            "Which Binding do you want to use?",
            default=1,
            choices=[str(i) for i in range(len(bindings))],
        )

        with console.status("Advertising beacon"):
            asyncio.run(
                advertise(
                    bindings[int(bind_index)],
                    [AdvertiseBeacon(url=url)],
                    interval=interval,
                    iterations=iterations,
                )
            )

    else:
        with console.status("Advertising beacons"):
            asyncio.run(
                adversite_all(
                    bindings,
                    [AdvertiseBeacon(url=url)],
                    interval=interval,
                    iterations=iterations,
                )
            )


if __name__ == "__main__":
    cli()



import yaml
from fakts_next.protocols import FaktValue
from typing import Dict
from pydantic import BaseModel, ConfigDict
from fakts_next.grants.base import BaseFaktsGrant


class YamlGrant(BaseFaktsGrant):
    """
    Represent a Grant that loads configuration from a YAML file.

    Attributes
    ----------
    filepath : str
        The path of the YAML file.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    filepath: str

    async def aload(self) -> Dict[str, FaktValue]:
        """Loads the YAML file and returns the configuration."""
        with open(self.filepath, "r") as file:
            config = yaml.load(file, Loader=yaml.FullLoader)  # type: ignore #TODO: Check why this is not working

        return config



from fakts_next.fakts import get_current_fakts_next
from koil.helpers import unkoil
from .protocols import FaktValue


async def afakt(key: str) -> FaktValue:
    return await get_current_fakts_next().aget(key)


def fakt(key: str) -> FaktValue:
    return unkoil(afakt, key)



""" Fakts grants module.

This module contains the grants that are included with
the fakts_next package.

Grants are the way to get configuration from different
sources. They are the main way to get configuration, and
are used by the Fakts class to get the configuration.

Generally, the grants are split into a few categories:

- IO grants: These grants are used to read configuration
    from files, and are generally used to read static
    configuration from files.
- Remote grants: These grants are used to connect to
    remote configuration servers, and fetch the configuration
    from there. They are generally used to fetch dynamic
    configuration from a remote server.
- Meta grants: These grants are used to combine multiple
    grants together, and are generally used to combine
    multiple grants together, to get configuration from
    multiple sources at the same time.
- Save grants: These grants are used to  load
    the configuration from another grant and cache it
    on disk, so that it can be used later without
    having to fetch it again.
- Env grants: These grants are used to load configuration
    from environment variables.


"""

from .env import EnvGrant
from .errors import GrantError
from .meta import FailsafeGrant, ParallelGrant
from .remote import RemoteGrant


__all__ = [
    "EnvGrant",
    "GrantError",
    "FaktsGrant",
    "YamlGrant",
    "FailsafeGrant",
    "ParallelGrant",
    "RemoteGrant",
]



try:
    import tomllib as toml
except ImportError:
    try:
        import toml  # type: ignore
    except ImportError:
        raise ImportError(
            "You need to install the `toml` package to use the toml grant."
            "Or use python 3.11 and higher which comes with the `tomllib` package"
        )


from fakts_next.grants.base import BaseFaktsGrant
from fakts_next.protocols import FaktsRequest
from fakts_next.protocols import FaktValue
from typing import Dict


class TomlGrant(BaseFaktsGrant):
    """
    A class used to represent a Grant in a TOML file.

    Attributes
    ----------
    filepath : str
        a formatted string to print out the file path where the TOML file is located

    Methods
    -------
    aload(request: FaktsRequest)
        Asynchronously loads the TOML file and returns the configuration.
    """

    filepath: str

    async def aload(self, request: FaktsRequest) -> Dict[str, FaktValue]:
        """Loads the TOML file and returns the configuration.

        Parameters
        ----------
        request : FaktsRequest
            The request object that may contain additional information needed for loading the TOML file.

        Returns
        -------
        dict
            The configuration loaded from the TOML file.
        """
        with open(self.filepath, "r") as file:
            config = toml.load(file)

        return config



from pydantic import ConfigDict, Field
from fakts_next.grants.errors import GrantError
import yaml
from koil.qt import QtCoro, QtFuture
from qtpy import QtWidgets
from fakts_next.protocols import FaktValue
from typing import Dict, Protocol, runtime_checkable, Optional, Type
from pydantic import BaseModel


class NoFileSelected(GrantError):
    """Raised when no file was selected."""

    pass


class QtSelectYaml(QtWidgets.QFileDialog):
    """
    Represents a File Dialog that selects YAML files.

    Methods
    -------
    ask(parent=None)
        Opens the file dialog and returns the selected file path.
    """

    def __init__(self, *args, **kwargs) -> None:
        """A File Dialog that selects YAML files."""
        super().__init__(*args, **kwargs)
        self.setFileMode(QtWidgets.QFileDialog.ExistingFile)
        self.setNameFilter("YAML files (*.yaml)")

    @classmethod
    def ask(
        cls: Type["QtSelectYaml"], parent: Optional[QtWidgets.QWidget] = None
    ) -> str:
        """
        Opens the file dialog and returns the selected file path.

        Parameters
        ----------
        parent : QWidget, optional
            The parent widget of the file dialog.

        Returns
        -------
        str
            The file path of the selected file.
        """
        filepath, weird = cls.getOpenFileName(parent=parent, caption="Select a Yaml")
        return filepath


class WrappingWidget(QtWidgets.QWidget):
    """
    A Widget that wraps the file selection process.

    Attributes
    ----------
    get_file_coro : QtCoro
        The coroutine that opens the file dialog and returns the selected file path
        (a call to `get_file_coro.acall()` will return a future that resolves to the selected file path
        or rejects with a `NoFileSelected` error).

    Methods
    -------
    open_file(future: QtFuture)
        Opens the file dialog and resolves or rejects the future based on the selected file path.
    """

    def __init__(self, *args, **kwargs) -> None:
        """A Widget that wraps the file selection process."""
        super().__init__(*args, **kwargs)
        self.get_file_coro = QtCoro(self.open_file)

    def open_file(self, future: QtFuture) -> None:
        """
        Opens the file dialog and resolves or rejects the future based on the selected file path.

        Parameters
        ----------
        future : QtFuture
            The future that will be resolved or rejected based on the selected file path.
        """
        filepath = QtSelectYaml.ask(parent=self)

        if filepath:
            future.resolve(filepath)
        else:
            future.reject(NoFileSelected("No file selected"))

    async def aask(self) -> str:
        """
        Opens the file dialog and returns the selected file path.

        Returns
        -------
        str
            The file path of the selected file.
        """
        return await self.get_file_coro.acall()


@runtime_checkable
class FileWidget(Protocol):
    """A Protocol that represents a widget

    that wraps the file selection process.
    It can be used to create a custom widget that wraps the file selection process.

    """

    async def aask(self) -> str:
        """Opens the file dialog and returns the selected file path.

        Parameters
        ----------
        parent : QWidget, optional
            The parent widget of the file dialog.

        Returns
        -------
        str
            The file path of the selected file.
        """
        ...


class QtYamlGrant(BaseModel):
    """
    Represent a Grant that allows the user to select a YAML file.

    Attributes
    ----------
    widget : WrappingWidget
        The widget that wraps the file selection process.

    Methods
    -------
    aload(request: FaktsRequest)
        Asynchronously loads the YAML file and returns the configuration.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    widget: FileWidget = Field(exclude=True)

    async def aload(self) -> Dict[str, FaktValue]:
        """Loads the YAML file and returns the configuration.

        Parameters
        ----------
        request : FaktsRequest
            The request object that may contain additional information needed for loading the YAML file.

        Returns
        -------
        dict
            The configuration loaded from the YAML file.
        """
        filepath = await self.widget.aask()
        with open(filepath, "r") as file:
            config = yaml.load(file, Loader=yaml.FullLoader)  # type: ignore #TODO: Check why this is not working

        return config



"""Meta grant Module

In this module you can find all the grants that are used to combine multiple grants together.

"""

from .failsafe import FailsafeGrant
from .parallel import ParallelGrant

__all__ = ["FailsafeGrant", "ParallelGrant"]



from fakts_next.grants.remote.discovery.advertised import (
    alisten_pure,
    ListenBinding,
    Beacon,
)
from qtpy import QtWidgets, QtCore, QtGui
import asyncio
import logging
from koil.qt import QtCoro, QtFuture, QtSignal
from fakts_next.grants.remote.discovery.utils import discover_url
from pydantic import BaseModel, ConfigDict
from typing import Any, Optional
from typing import Dict, List

from pydantic import Field
import ssl
import certifi
from fakts_next.grants.remote.models import FaktsEndpoint


logger = logging.getLogger(__name__)


class SelfScanWidget(QtWidgets.QWidget):
    """A widget that allows the user to scan for beacons

        It has a line edit and a button. When the button is clicked,
        the line edit is scanned for a url, and if it is a valid url,
        a beacon is emitted with the url.
    from .utils import discover_url

        No validation is done on the url, so it is up to the user to
        enter a valid url.

    """

    user_beacon_added = QtCore.Signal(Beacon)

    def __init__(self, *args, **kwargs) -> None:
        """Constructor for SelfScanWidget"""
        super().__init__(*args, **kwargs)
        self.scanlayout = QtWidgets.QHBoxLayout()
        self.lineEdit = QtWidgets.QLineEdit()
        self.addButton = QtWidgets.QPushButton("Scan")

        self.scanlayout.addWidget(self.lineEdit)
        self.scanlayout.addWidget(self.addButton)
        self.addButton.clicked.connect(self.on_add)
        self.setLayout(self.scanlayout)

    def on_add(self) -> None:
        """Called when the button is clicked"""
        host = self.lineEdit.text()
        beacon = Beacon(url=host)
        self.user_beacon_added.emit(beacon)


class FaktsEndpointButton(QtWidgets.QPushButton):
    """A button that represents a FaktsEndpoint

    It has a title and a subtitle, and when clicked, emits
    a signal with the endpoint
    """

    accept_clicked = QtCore.Signal(FaktsEndpoint)

    def __init__(
        self, endpoint: FaktsEndpoint, parent: Optional[QtWidgets.QWidget] = None
    ) -> None:
        """Constructor for FaktsEndpointButton"""
        super(FaktsEndpointButton, self).__init__(parent)
        self.endpoint = endpoint
        self.hovered = False
        self.is_pressed = False
        self.clicked.connect(self.on_clicked)
        self.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))  # type: ignore  # Default cursor
        self.initUI()

    def initUI(self) -> None:
        """Initialize UI elements"""
        # Set the button's minimum height and other properties if needed
        self.setMinimumHeight(50)
        self.setContentsMargins(10, 10, 10, 10)
        self.setToolTip("Connect to " + self.endpoint.name)

    def paintEvent(self, event: Any) -> None:
        """Paint the button with the given style, title and subtitle"""
        # Call the superclass paint event to keep the button look normal
        super(FaktsEndpointButton, self).paintEvent(event)

        painter = QtGui.QPainter(self)
        painter.setRenderHint(QtGui.QPainter.Antialiasing)

        # Calculate the position for the title and subtitle
        title_pos_y = int(self.rect().height() / 3)
        subtitle_pos_y = int(self.rect().height() / 3 * 2)

        # Set the font for the title
        title_font = QtGui.QFont("SansSerif", 12)  # Make the title font a bit bigger
        title_font.setBold(True)
        painter.setFont(title_font)
        painter.drawText(
            QtCore.QRect(10, 10, self.width(), title_pos_y),
            QtCore.Qt.AlignLeft,  # type: ignore
            self.endpoint.name,
        )

        # Set the font for the subtitle
        subtitle_font = QtGui.QFont("SansSerif", 8)
        painter.setFont(subtitle_font)
        painter.drawText(
            QtCore.QRect(10, title_pos_y + 10, self.width(), subtitle_pos_y),
            QtCore.Qt.AlignLeft,  # type: ignore
            "@" + self.endpoint.base_url,
        )

    def enterEvent(self, event: Any) -> None:
        """Sets the cursor to pointing hand when hovering over the button"""
        self.hovered = True

        self.update()  # Trigger a repaint
        self.setCursor(
            QtGui.QCursor(QtCore.Qt.PointingHandCursor)  # type: ignore
        )  # Change cursor to pointer

    def leaveEvent(self, event: Any) -> None:
        """Sets the cursor to arrow when not hovering over the button"""
        self.hovered = False
        self.setCursor(QtGui.QCursor(QtCore.Qt.ArrowCursor))  # type: ignore  # Change cursor to arrow
        self.update()  # Trigger a repaint

    def mousePressEvent(self, event: Any) -> None:
        """Sets the button to pressed when clicked"""
        self.is_pressed = True
        self.update()  # Trigger a repaint
        super(FaktsEndpointButton, self).mousePressEvent(event)

    def mouseReleaseEvent(self, event: Any) -> None:
        """Sets the button to not pressed when released"""
        self.is_pressed = False
        self.update()  # Trigger a repaint
        super(FaktsEndpointButton, self).mouseReleaseEvent(event)

    def on_clicked(self) -> None:
        """Called when the button is clicked"""
        self.accept_clicked.emit(self.endpoint)


class SelectBeaconWidget(QtWidgets.QDialog):
    """A widget that allows the user to select a beacon

    It has a list of buttons, each representing a beacon.
    When a button is clicked, a signal is emitted with the beacon
    """

    new_advertised_endpoint = QtCore.Signal(FaktsEndpoint)
    new_local_endpoint = QtCore.Signal(FaktsEndpoint)

    def __init__(
        self, *args, settings: Optional[QtCore.QSettings] = None, **kwargs
    ) -> None:
        """Constructor for SelectBeaconWidget"""
        super().__init__(*args, **kwargs)
        self.setWindowTitle("Search Endpoints...")
        self.hide_coro = QtCoro(self.hide, autoresolve=True)
        self.show_error_coro = QtCoro(self.show_error, autoresolve=True)
        self.clear_endpoints_coro = QtCoro(self.clear_endpoints, autoresolve=True)
        self.select_endpoint = QtCoro(self.demand_selection_of_endpoint)
        self.settings = settings

        self.select_endpoint_future = None

        self.new_advertised_endpoint.connect(self.on_new_endpoint)
        self.new_local_endpoint.connect(self.on_new_endpoint)

        self.endpoints: List[FaktsEndpoint] = []

        self.wlayout = QtWidgets.QVBoxLayout()

        self.endpointLayout = QtWidgets.QVBoxLayout()

        self.scanWidget = SelfScanWidget()
        self.beacon_user = QtSignal(self.scanWidget.user_beacon_added)

        QBtn = QtWidgets.QDialogButtonBox.Cancel
        self.buttonBox = QtWidgets.QDialogButtonBox(QBtn)
        self.buttonBox.rejected.connect(self.on_reject)

        endgroup = QtWidgets.QGroupBox("Select")
        endgroup.setLayout(self.endpointLayout)
        endgroup.setStyleSheet("padding: 10px;")

        self.scan_layout = QtWidgets.QVBoxLayout()
        self.scan_layout.addWidget(self.scanWidget)

        scangroup = QtWidgets.QGroupBox("Or Scan")
        scangroup.setLayout(self.scan_layout)

        self.wlayout.addWidget(endgroup)
        self.wlayout.addWidget(scangroup)
        self.wlayout.addWidget(self.buttonBox)
        self.setLayout(self.wlayout)

    def clearLayout(self, layout: QtWidgets.QVBoxLayout) -> None:
        """Clear the layout"""
        while layout.count():
            child = layout.takeAt(0)
            if child.widget():
                child.widget().deleteLater()

    def clear_endpoints(self) -> None:
        """Clear the endpoints"""
        self.clearLayout(self.endpointLayout)
        self.endpoints = []

    def show_me(self) -> None:
        """A function that shows the widget"""
        self.show()

    def show_error(self, error: Exception) -> None:
        """Show an error message

        Parameters
        ----------
        error : Exception
            Display the error message
        """
        self.show()
        QtWidgets.QMessageBox.critical(self, "Error", str(error))

    def demand_selection_of_endpoint(self, future: QtFuture) -> None:
        """Is called when the user should select an endpoint. I.e. when the demander
        is called programmatically"""
        self.select_endpoint_future = future
        self.show()

    def on_endpoint_clicked(self, item: FaktsEndpoint) -> None:
        """Called when an endpoint button is clicked

        Will resolve the future with the endpoint



        """
        if self.select_endpoint_future:
            self.select_endpoint_future.resolve(item)

    def on_reject(self) -> None:
        """Called when the user rejects the dialog

        Will reject the future"""
        if self.select_endpoint_future:
            self.select_endpoint_future.reject(
                Exception("User cancelled the this Grant without selecting a Beacon")
            )
        self.reject()

    def closeEvent(self, event: Any) -> None:
        """Called when the window is closed. Will automatically reject the future"""
        if self.select_endpoint_future:
            self.select_endpoint_future.reject(
                Exception("User cancelled the this Grant without selecting a Beacon")
            )

        event.accept()  # let the window close

    def on_new_endpoint(self, config: FaktsEndpoint) -> None:
        """A callback that is called when a new endpoint is discovered

        Creates a button for the endpoint and adds it to the layout


        """
        self.clearLayout(self.endpointLayout)

        self.endpoints.append(config)

        for endpoint in self.endpoints:
            widget = FaktsEndpointButton(endpoint)

            self.endpointLayout.addWidget(widget)
            widget.accept_clicked.connect(self.on_endpoint_clicked)


async def wait_first(*tasks) -> Any:
    """Return the result of first async task to complete with a non-null result"""
    # Get first completed task(s)
    done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)

    # Cancel pending tasks
    for task in pending:
        task.cancel()

    # Wait for pending tasks to be cancelled
    await asyncio.gather(*pending, return_exceptions=True)

    # Return first completed task result
    for task in done:
        return task.result()


class QtSelectableDiscovery(BaseModel):
    """A QT based discovery that will cause the user to select an endpoint
    from a list of network discovered endpoints, as well as endpoints
    that the user manually enters.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    binding: ListenBinding = Field(default_factory=ListenBinding)
    """The address to bind to"""
    strict: bool = False
    """Should we error on bad Beacons"""
    discovered_endpoints: Dict[str, FaktsEndpoint] = Field(default_factory=dict)
    """A cache of discovered endpoints"""
    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""
    allow_appending_slash: bool = Field(
        default=True,
        description="If the url does not end with a slash, should we append one? ",
    )
    auto_protocols: List[str] = Field(
        default_factory=lambda: [],
        description="If no protocol is specified, we will try to connect to the following protocols",
    )
    timeout: int = Field(
        default=3,
        description="The timeout for the connection",
    )
    additional_beacons: List[str] = Field(
        default_factory=lambda: ["localhost:11000", "localhost:11001", "localhost:8000"]
    )
    widget: SelectBeaconWidget

    async def emit_endpoints(self) -> None:
        """A long running task that will emit endpoints that are discovered
        through the network and emit them to be displayed in the widget.
        """
        if self.additional_beacons:
            for i in self.additional_beacons:
                try:
                    endpoint = await discover_url(
                        i,
                        self.ssl_context,
                        auto_protocols=self.auto_protocols,
                        allow_appending_slash=self.allow_appending_slash,
                        timeout=self.timeout,
                    )
                    self.widget.new_local_endpoint.emit(endpoint)
                except Exception as e:
                    logger.info(f"Could not connect to localhost: {e}")

        try:
            try:
                binding = self.binding
                async for beacon in alisten_pure(binding, strict=self.strict):
                    try:
                        if beacon.url in self.additional_beacons:
                            # we already did this one
                            continue
                        endpoint = await discover_url(
                            beacon.url,
                            self.ssl_context,
                            auto_protocols=self.auto_protocols,
                            allow_appending_slash=self.allow_appending_slash,
                            timeout=self.timeout,
                        )
                        self.widget.new_advertised_endpoint.emit(endpoint)
                    except Exception as e:
                        logger.info(f"Could not connect to beacon: {beacon.url} {e}")
            except Exception:
                logger.exception("Error in discovery")
                return None

        except Exception as e:
            logger.exception(e)
            raise e

    async def await_user_definition(
        self,
    ) -> Optional[FaktsEndpoint]:
        """Await the user to define a beacon. This will cause the widget to listen for
        user input.

        Returns
        -------
        FaktsEndpoint
            The endpoint that the user defined

        """

        async for beacon in self.widget.beacon_user.aiterate():
            try:
                return await discover_url(
                    beacon.url,
                    self.ssl_context,
                    auto_protocols=self.auto_protocols,
                    allow_appending_slash=self.allow_appending_slash,
                    timeout=self.timeout,
                )
            except Exception as e:
                await self.widget.show_error_coro.acall(e)
                logger.error(f"Could not connect to beacon: {beacon.url} {e}")
                continue

        return None

    async def adiscover(self) -> FaktsEndpoint:
        """Discover the endpoint

        This discovery will cause the attached widget to be shown, and
        then the user is able to select and endpoint that is discovered
        automatically or by the user.

        Parameters
        ----------
        request : FaktsRequest
            The request to use for the discovery process (is not used)

        Returns
        -------
        FaktsEndpoint
            A valid endpoint
        """

        emitting_task = asyncio.create_task(self.emit_endpoints())
        try:
            await self.widget.clear_endpoints_coro.acall()

            try:
                select_endpoint_task = asyncio.create_task(
                    self.widget.select_endpoint.acall()
                )
                user_definition_task = asyncio.create_task(self.await_user_definition())

                endpoint: FaktsEndpoint = await wait_first(
                    select_endpoint_task, user_definition_task
                )

                await self.widget.hide_coro.acall()

            finally:
                emitting_task.cancel()
                try:
                    await emitting_task
                except asyncio.CancelledError:
                    logger.info("Cancelled the Discovery task")

            return endpoint
        except Exception as e:
            logger.exception(e)
            emitting_task.cancel()

            try:
                await emitting_task
            except asyncio.CancelledError:
                logger.info("Cancelled the Discovery task")

            raise e



""" Fakts CLI module.

This module contains the CLI for the fakts_next package.
It allows you act and interface with the fakts_next package
from the command line.

Currently it only supports advertising a fakts_next endpoint
on the local network, but in the future it will also
support other features, like fetching the configuration
from a remote endpoint.

"""



from typing import List, Dict
import asyncio
from functools import reduce

from fakts_next.utils import update_nested
from fakts_next.protocols import FaktValue, FaktsGrant
from pydantic import BaseModel, ConfigDict


class ParallelGrant(BaseModel):
    """A grant that loads multiple grants in parallel and merges the results


    Attributes
    ----------
    grants : List[FaktsGrant]
        The grants to load in parallel

    omit_exceptions : bool
        Omit exceptions if any of the grants fail to load (otherwise will raise an exception)

    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    grants: List[FaktsGrant]
    " The grants to load in parallel "
    omit_exceptions: bool = False
    " Omit exceptions if any of the grants fail to load "

    async def aload(self) -> Dict[str, FaktValue]:
        """Loads the configuration from the grants in parallel

        This method will load the configuration from the grants in parallel, and merge the results, updating
        the configuration with the results from each grant.

        Parameters
        ----------
        request : FaktsRequest
            The request object that may contain additional information needed for loading the configuration.

        Returns
        -------
        dict
            The configuration loaded from the grant.


        """
        config_futures = [grant.aload() for grant in self.grants]
        configs = await asyncio.gather(
            *config_futures, return_exceptions=self.omit_exceptions
        )
        # TODO: Check if this is the correct way to merge the configs
        return reduce(lambda x, y: update_nested(x, y), configs, {})  # type: ignore

    async def arefresh(self):
        config_futures = [grant.arefresh() for grant in self.grants]
        configs = await asyncio.gather(
            *config_futures, return_exceptions=self.omit_exceptions
        )
        # TODO: Check if this is the correct way to merge the configs
        return reduce(lambda x, y: update_nested(x, y), configs, {})  # type: ignore



from typing import List, Dict
from fakts_next.grants.errors import GrantError
from fakts_next.grants.base import BaseFaktsGrant
import logging
from fakts_next.protocols import FaktsGrant, FaktValue
from pydantic import BaseModel, ConfigDict

logger = logging.getLogger(__name__)


class FailsafeGrant(BaseFaktsGrant):
    """
    Represent a Grant that loads configuration from a selection
    of other grants. It will try to load the grants in order,
    and will return the values from the first grant that succeeds.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    grants: List[FaktsGrant]

    async def aload(self) -> Dict[str, FaktValue]:
        """Loads the configuration from the grant

        It will try to load the grants in order, and will return the values from the first grant that succeeds.


        Parameters
        ----------
        request : FaktsRequest
            The request object that may contain additional information needed for loading the configuration.

        Returns
        -------
        dict
            The configuration loaded from the grant.


        """
        for grant in self.grants:
            try:
                config = await grant.aload()
                return config
            except GrantError:
                logger.exception(f"Failed to load {grant}", exc_info=True)
                continue

        raise GrantError("Failed to load any grants")

    async def arefresh(self) -> Dict[str, FaktValue]:
        """Loads the configuration from the grant

        It will try to load the grants in order, and will return the values from the first grant that succeeds.


        Parameters
        ----------
        request : FaktsRequest
            The request object that may contain additional information needed for loading the configuration.

        Returns
        -------
        dict
            The configuration loaded from the grant.


        """
        for grant in self.grants:
            try:
                config = await grant.arefresh()
                return config
            except GrantError:
                logger.exception(f"Failed to load {grant}", exc_info=True)
                continue

        raise GrantError("Failed to load any grants")



import asyncio
import contextvars
import logging
from typing import Any, Dict

from koil.composition import KoiledModel
from koil.helpers import unkoil
from pydantic import Field
from typing import Optional
from fakts_next.errors import GroupNotFound, NoFaktsFound
from fakts_next.cache.nocache import NoCache
from .protocols import FaktsCache, FaktValue, FaktsGrant

logger = logging.getLogger(__name__)
current_fakts_next: contextvars.ContextVar[Optional["Fakts"]] = contextvars.ContextVar(
    "current_fakts_next", default=None
)


class Fakts(KoiledModel):
    """Fakts is any asynchronous configuration loader.

    Fakts provides a way to concurrently load and access configuration from different
    sources in async and sync environments.

    It is used to load configuration from a grant, and to access it in async
    and sync code.

    A grant constitutes the way to load configuration. It can be a local config file
    (eg. yaml, toml, json), environemnt variables, a remote configuration (eg. from
    a fakts server) a database, or any other source.  It will be loaded either on
    call to `load`,  or on  a call to `get` (if auto_load is set to true).

    Additionaly you can compose grants with the help of meta grants in order to
    load configuration from multiple sources.

    Example:
        ```python
        async with Fakts(grant=YamlGrant("config.yaml")) as fakts:
            config = await fakts.aget("group_name")
        ```

        or

        ```python
        with Fakts(grant=YamlGrant("config.yaml")) as fakts:
            config = await fakts.get("group_name")
        ```

    Fakts should be used as a context manager, and will set the current fakts context
    variable to itself, letting you access the current fakts instance from anywhere in
    your code (async or sync). To understand how the async sync code access work,
    please check out the documentation for koil.


    Example:
        ```python
        async with Fakts(grant=FailsafeGrant(
            grants=[
                EnvGrant(),
                YamlGrant("config.yaml")
            ]
        )) as fakts:
            config = await fakts.get("group_name")
        ```
        In this example fakts will load the configuration from the environment
        variables first, and if that fails, it will load it from the yaml file.


    """

    cache: FaktsCache = Field(default_factory=NoCache, exclude=True)
    grant: FaktsGrant
    """The grant to load the configuration from"""

    hard_fakts: Dict[str, Any] = Field(default_factory=dict, exclude=True)
    """Hard fakts are fakts that are set by the user and cannot be overwritten by grants"""

    loaded_fakts: Optional[Dict[str, Any]] = Field(default=None, exclude=True)
    """The currently loaded fakts. Please use `get` to access the fakts"""

    allow_auto_load: bool = Field(
        default=True, description="Should we autoload on get?"
    )
    """Should we autoload the grants on a call to get?"""

    load_on_enter: bool = False
    """Should we load on connect?"""
    delete_on_exit: bool = False
    """Should we delete on connect?"""

    refetch_on_group_not_found: bool = False

    _loaded: bool = False
    _lock: Optional[asyncio.Lock] = None

    async def aget(
        self,
        group_name: Optional[str] = None,
        **kwargs,
    ) -> FaktValue:
        """Get Fakt Value (async)

        Gets the currently active configuration for the group_name, by loading it from
        the grant if it is not already loaded.

        Steps:
            1. Acquire lock
            2. If not yet loaded and auto_load is True, load
            4. Return groups fakts

        Args:
            group_name (str): The group name in the fakts
            auto_load (bool, optional): Should we autoload the configuration
                                        if nothing has been set? Defaults to True.
            force_refresh (bool, optional): Should we force a refresh of the grants.
                                            Grants can decide their own refresh logic?
                                            Defaults to False.

        Returns:
            dict: The active fakts
        """
        assert (
            self._lock is not None
        ), "You need to enter the context first before calling this function"
        async with self._lock:
            if not self.loaded_fakts:
                try:
                    await self.aload()
                except Exception as e:
                    logger.error(e, exc_info=True)
                    raise e

        try:
            config = self._getsubgroup(group_name, base=group_name)
        except GroupNotFound as e:
            if self.refetch_on_group_not_found:
                try:
                    await self.arefresh()
                except Exception as e:
                    logger.error(e, exc_info=True)
                    raise e

                try:
                    config = self._getsubgroup(group_name, base=group_name)
                except GroupNotFound as e:
                    raise GroupNotFound(
                        f"Could't find {group_name} in fakts. Even after refresh"
                    ) from e
            else:
                raise e

        return config

    def _getsubgroup(
        self, group_name: Optional[str] = None, base: str = ""
    ) -> Dict[str, Any]:
        """Get subgroup

        Protected function to get a subgroup from the loaded fakts

        Args:
            group_name (str): The name of the group

        Raises:
            GroupNotFound: If the groups is not found in the loadedfakts

        Returns:
            Dict[str, Any]: The subgroups configuration as a dictioniary
        """
        config = {**self.loaded_fakts}

        if group_name is None:
            return config

        path = []

        for subgroup in group_name.split("."):
            try:
                path += [subgroup]
                config = config[subgroup]
            except KeyError as e:
                raise GroupNotFound(
                    f"Could't find {subgroup} in subgroup when trying to access '"
                    + " > ".join(path)
                    + f"'. Available keys in this config are {', '.join(list(config.keys()))}"
                ) from e

        return config

    def has_changed(self, value: FaktValue, group: str) -> bool:
        """Has Changed

        Checks if the value has changed since the last load.


        Parameters
        ----------
        value: FaktValue
            The value to check
        group : str
            The group it belongs to

        Returns
        -------
        bool
            True if the value has changed, False otherwise
        """

        return (
            not value or self._getsubgroup(group) != value
        )  # TODO: Implement Hashing on config?

    async def arefresh(self, **kwargs) -> Dict[str, Any]:
        """Causes a Refresh, by reloading the grants"""
        self.loaded_fakts = {}
        await self.cache.areset()
        return await self.aload()

    def refresh(self, **kwargs) -> Dict[str, Any]:
        """Causes a Refresh, by reloading the grants"""
        return unkoil(self.arefresh, **kwargs)

    def get(
        self,
        group_name: Optional[str] = None,
        **kwargs,
    ) -> FaktValue:
        """Get Fakt Value (Sync)

        Gets the currently active configuration for the group_name, by loading it from
        the grant if it is not already loaded.

        Steps:
            1. Acquire lock
            2. If not yet loaded and auto_load is True, load
            4. Return groups fakts

        Args:
            group_name (str): The group name in the fakts
            auto_load (bool, optional): Should we autoload the configuration
                                        if nothing has been set? Defaults to True.
            force_refresh (bool, optional): Should we force a refresh of the grants.
                                            Grants can decide their own refresh logic?
                                            Defaults to False.

        Returns:
            dict: The active fakts
        """
        return unkoil(self.aget, group_name=group_name, **kwargs)

    async def aload(self) -> Dict[str, FaktValue]:
        """Loads the configuration from the grant (async)

        This method will load the configuration from the grant, and set it as the
        the currently active configuration. It is called automatically on a call to
        `get` if the configuration has not been loaded yet.

        In contrary to `refresh`, a potential cached configuration can be used, if


        Parameters
        ----------
        request : FaktsRequest
            The request that is being processed.

        Returns
        -------

        Dict[str, FaktValue]
           The loaded fakts


        """
        try:
            self.loaded_fakts = await self.cache.aload()
        except Exception as e:
            logger.error(e, exc_info=True)
            raise e

        # Cache is empty, we need to reload the grants
        if not self.loaded_fakts:
            try:
                self.loaded_fakts = await self.grant.aload()
                await self.cache.aset(self.loaded_fakts)
            except Exception as e:
                logger.error(e, exc_info=True)
                raise e

        return self.loaded_fakts

    async def arefresh(self) -> Dict[str, FaktValue]:

        self.loaded_fakts = None
        await self.cache.areset()
        return await self.aload()

    def load(self) -> Dict[str, FaktValue]:
        """Loads the configuration from the grant (sync)

        This method will load the configuration from the grant, and set it as the
        the currently active configuration. It is called automatically on a call to
        `get` if the configuration has not been loaded yet.

        Parameters
        ----------
        request : FaktsRequest
            The request that is being processed.

        Returns
        -------

        Dict[str, FaktValue]
           The loaded fakts


        """
        return unkoil(self.aload)

    async def __aenter__(self) -> "Fakts":
        """Enter the context manager

        This method will set the current fakts context variable to itself,
        and create locks, to make sure that only one fakt request is
        processed at a time.
        """

        current_fakts_next.set(
            self
        )  # TODO: We should set tokens, but depending on async/sync this is shit
        self._lock = asyncio.Lock()
        return self

    async def __aexit__(self, *args, **kwargs) -> None:
        """Exit the context manager and clean up"""
        current_fakts_next.set(
            None
        )  # TODO: And here we should reset, but can't because of koil unsafe thread

    def _repr_html_inline_(self) -> str:
        """(Internal) HTML representation for jupyter"""
        return f"<table><tr><td>grant</td><td>{self.grant.__class__.__name__}</td></tr></table>"


def get_current_fakts_next() -> Fakts:
    """Get the current fakts instance

    This method will return the current fakts instance, or raise an
    exception if no fakts instance is set.

    Returns
    -------
    Fakts
        The current fakts instance
    """
    fakts = current_fakts_next.get()

    if fakts is None:
        raise NoFaktsFound("No fakts instance set in this context")

    return fakts
